/*    Copyright (C) 2010  Tomasz Bursztyka
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <manage_cmd.h>

extern in_plugin_handlers_t *ph;

extern char mng_password[MNG_MAX_PASSWORD];

int manage_cmd(int argc, char **argv)
{
	uint16_t port = MNG_PORT_DEFAULT;

	if ((argc < 2) || (argc > 3))
	{
		log_msg(ERROR_LVL, "manage command takes 1 or 2 arguments.\n");
		return CMD_ERROR;
	}

	if (strlen(argv[1]) >= MNG_MAX_PASSWORD)
	{
		log_msg(ERROR_LVL, "manage password cannot be longer than %d\n", MNG_MAX_PASSWORD - 1);
		return CMD_ERROR;
	}

	memset(mng_password, 0, MNG_MAX_PASSWORD);
	strncpy(mng_password, argv[1], MNG_MAX_PASSWORD);

	if (argc == 3)
	{
		port = (uint16_t) strtol(argv[2], (char **)NULL, 10);
		if ((errno == ERANGE) && ((port == USHRT_MAX) || (port == 0)))
		{
			log_msg(ERROR_LVL, "manage port: wrong number - using default one\n");
			port = MNG_PORT_DEFAULT;
		}
	}

	DEBUG_MSG("Manage password %s port %d\n", mng_password, port);

	if (setup_mng_socket(port) < 0)
	{
		log_msg(ERROR_LVL, "manage interface could not start.\n");
		return CMD_ERROR;
	}

	return CMD_SUCCESS;
}

int quit_cmd(int argc, char **argv)
{
	argc = 0;
	argv = NULL;

	client_disconnect();

	return CMD_SUCCESS;
}

int disconnect_cmd(int argc, char **argv)
{
	struct in6_addr ip_mnode;
	mnode_t *mnode = NULL;

	if (argc != 2)
	{
		log_msg(ERROR_LVL, "list command takes 1 argument (and only one).\n");
		return CMD_ERROR;
	}

	if (inet_pton(AF_INET6, argv[1], &ip_mnode) <= 0)
	{
		log_msg(ERROR_LVL, "disconnect command: invalid ip address %s\n", argv[1]);
		return CMD_ERROR;
	}

	mnode = h_search(mdevice_g->ip_mnodes, &ip_mnode);
	if (mnode != NULL)
		remove_mnode(mdevice_g, mnode);
	else
		log_msg(INFO_LVL, "disconnect %s: no such ip.\n", argv[1]);

	return CMD_SUCCESS;
}

int list_cmd(int argc, char **argv)
{
	mnode_t *mnode = NULL;
	char ip_str[INET6_ADDRSTRLEN];
	char uip_str[INET6_ADDRSTRLEN];
	int idx = 0;
	mserver_t *mserver = NULL;

	if (argc != 2)
	{
		log_msg(ERROR_LVL, "list command takes 1 argument (and only one).\n");
		return CMD_ERROR;
	}

	if (strncmp(argv[1], "mnodes", 7) == 0)
	{
		if ((mdevice_g->ip_mnodes == NULL) ||
			(h_count(mdevice_g->ip_mnodes) == 0))
		{
			log_msg(INFO_LVL, "There is no mnodes on this vpzone instance\n");
			return CMD_SUCCESS;
		}
	
		h_iter_begin(mdevice_g->ip_mnodes);

		log_msg(INFO_LVL, "mnode host:port : vpn link IP - vpn global IP (received bytes / send bytes)\n");
		while(h_iter_next(mdevice_g->ip_mnodes))
		{
			if (IN6_IS_ADDR_LINKLOCAL((struct in6_addr *)h_iter_getkey(mdevice_g->ip_mnodes)))
			{
				mnode = h_iter_getvalue(mdevice_g->ip_mnodes);
				log_msg(INFO_LVL, "mnode %s:%d : %s - %s (%llu / %llu)\n", 
						mnode->host, mnode->port,
						inet_ntop(AF_INET6, &(mnode->vpn_uip), uip_str, INET6_ADDRSTRLEN),
						inet_ntop(AF_INET6, &(mnode->vpn_ip), ip_str, INET6_ADDRSTRLEN),
					    mnode->bytes_recv, mnode->bytes_send);
			}
		}

		h_iter_end(mdevice_g->ip_mnodes);

		return CMD_SUCCESS;
	}

	if (strncmp(argv[1], "mservers", 10) == 0)
	{
		if (mdevice_g->mservers != NULL)
		{
			mserver = mdevice_g->mservers;
			while(mserver != NULL)
			{
				log_msg(INFO_LVL, "mserver running on %s:%d\n",
						mserver->host, mserver->port);

				mserver = mserver->next;
			}
		}
		else
			log_msg(INFO_LVL, "There is no mservers running on this vpzone instance\n");

		return CMD_SUCCESS;
	}

	if (strncmp(argv[1], "connectors", 11) == 0)
	{
		for (idx = 0; idx < ph->connectors->length; idx++)
			if (ph->connectors->ptable[idx] != NULL)
				log_msg(INFO_LVL, "Connector type: %s\n", ((connector_t *) ph->connectors->ptable[idx])->type);

		if (idx == 0)
			log_msg(INFO_LVL, "No connectors.\n");

		return CMD_SUCCESS;
	}

	if (strncmp(argv[1], "ciphers", 8) == 0)
	{
		for (idx = 0; idx < ph->ciphers->length; idx++)
			if (ph->ciphers->ptable[idx] != NULL)
				log_msg(INFO_LVL, "Cipher type: %s\n", ((cipher_t *) ph->ciphers->ptable[idx])->type);

		if (idx == 0)
			log_msg(INFO_LVL, "No ciphers.\n");

		return CMD_SUCCESS;
	}

	if (strncmp(argv[1], "authenticators", 15) == 0)
	{
		for (idx = 0; idx < ph->authenticators->length; idx++)
			if (ph->authenticators->ptable[idx] != NULL)
				log_msg(INFO_LVL, "Authenticator type: %s\n", ((authenticator_t *) ph->authenticators->ptable[idx])->type);

		if (idx == 0)
			log_msg(INFO_LVL, "No authenticators.\n");

		return CMD_SUCCESS;
	}

	log_msg(ERROR_LVL, "list command: parameters %s unknown.\n", argv[1]);
	return CMD_ERROR;
}
