/*    Copyright (C) 2010  Tomasz Bursztyka
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <simple_vp.h>

static void send_vp_ips(mdevice_t *mdevice, mnode_t *mnode, enum simple_cmd_e cmd_num);

PLUGIN_DEPENDANCIES("tcp");

PLUGIN_GBL_VARS;

plugin_type_t init(int plugin_id, in_plugin_handlers_t *iph)
{
	PLUGIN_SET_VARS(plugin_id, iph);

	add_callback(CB_POST_CONNECT, init_simple_protocol);
	add_callback(CB_MANAGE_INTERNAL_PACKET, handle_simple_packet);

	log_msg(INFO_LVL, "Loaded SIMPLE_VP protocol plugin\n");

	return PLUGIN_DEFAULT;
}

int unload(void)
{
	sub_callback(CB_POST_CONNECT, init_simple_protocol);
	sub_callback(CB_MANAGE_INTERNAL_PACKET, handle_simple_packet);

	return 0;
}

void init_simple_protocol(callback_t *cb, mdevice_t *mdevice, mnode_t *mnode)
{
	DEBUG_MSG("SIMPLE = init protocol\n");

	send_vp_ips(mdevice, mnode, CMD_NEW);

	post_connect(cb->next, mdevice, mnode);
}

int handle_simple_packet(callback_t *cb, mdevice_t *mdevice, mnode_t *mnode)
{
	int pkt_len = 0;
	struct simple_vp_ips vp_ips;
	int ret;
#ifdef DEBUG
	char ip6_disp[INET6_ADDRSTRLEN];
	char vip6_disp[INET6_ADDRSTRLEN];
#endif

	if (SVP_PKT_ID(mnode->r_buffer) != SIMPLE_PACKET_ID)
	{
		DEBUG_MSG("SIMPLE = can't handle packet\n");
		manage_internal_packet(ret, cb->next, mdevice, mnode);
		return ret;
	}

	DEBUG_MSG("SIMPLE = handle packet\n");

	pkt_len = SVP_PKT_LEN(mnode->r_buffer) + SIMPLE_PACKET_HDR_LEN;

	if (pkt_len > mnode->r_bytes)
			return -1;

	switch (SVP_CMD_ID(mnode->r_buffer))
	{
	case CMD_NEW:
		DEBUG_MSG("SIMPLE = CMD_NEW\n");
		memcpy(&vp_ips, mnode->r_buffer+SIMPLE_PACKET_HDR_LEN, sizeof(struct simple_vp_ips));

		DEBUG_MSG("SIMPLE - Got vip6: %s and ip6: %s\n", 
				  inet_ntop(AF_INET6, &vp_ips.vp_uip, vip6_disp, INET6_ADDRSTRLEN),
				  inet_ntop(AF_INET6, &vp_ips.vp_ip, ip6_disp, INET6_ADDRSTRLEN));

		memcpy(&(mnode->vpn_uip), &(vp_ips.vp_uip), sizeof(struct in6_addr));
		memcpy(&(mnode->vpn_ip), &(vp_ips.vp_ip), sizeof(struct in6_addr));

		if (h_insert(mdevice->ip_mnodes, &(mnode->vpn_uip), mnode) != 0)
			DEBUG_MSG("Could not insert UIP\n");
		if (h_insert(mdevice->ip_mnodes, &(mnode->vpn_ip), mnode) != 0)
			DEBUG_MSG("Could not insert IP\n");

		send_vp_ips(mdevice, mnode, CMD_UPDATE);

		break;
	case CMD_UPDATE:
		DEBUG_MSG("SIMPLE = CMD_UPDATE\n");
		memcpy(&vp_ips, mnode->r_buffer+SIMPLE_PACKET_HDR_LEN, sizeof(struct simple_vp_ips));

		DEBUG_MSG("SIMPLE - Got vip6: %s and ip6: %s\n", 
				  inet_ntop(AF_INET6, &vp_ips.vp_uip, vip6_disp, INET6_ADDRSTRLEN),
				  inet_ntop(AF_INET6, &vp_ips.vp_ip, ip6_disp, INET6_ADDRSTRLEN));
		
		h_remove(mdevice->ip_mnodes, &(mnode->vpn_uip), false);
		h_remove(mdevice->ip_mnodes, &(mnode->vpn_ip), false);

		memcpy(&(mnode->vpn_uip), &(vp_ips.vp_uip), sizeof(struct in6_addr));
		memcpy(&(mnode->vpn_ip), &(vp_ips.vp_ip), sizeof(struct in6_addr));

		if (h_insert(mdevice->ip_mnodes, &(mnode->vpn_uip), mnode) != 0)
			DEBUG_MSG("Could not insert UIP\n");
		if (h_insert(mdevice->ip_mnodes, &(mnode->vpn_ip), mnode) != 0)
			DEBUG_MSG("Could not insert IP\n");

		break;
	default:
		break;
	}

	return pkt_len;
}

static void send_vp_ips(mdevice_t *mdevice, mnode_t *mnode, enum simple_cmd_e cmd_num)
{
	struct psimple_pkt_hdr pkt_hdr;
	struct simple_vp_ips vp_ips;
	uint8_t trans_buff[32768];
	int pkt_len;
#ifdef DEBUG
	char ip6_disp[INET6_ADDRSTRLEN];
	char vip6_disp[INET6_ADDRSTRLEN];
#endif

	DEBUG_MSG("SIMPLE = send vp ips\n");

	memset(trans_buff, 0, 32768);
	memset(&pkt_hdr, 0, SIMPLE_PACKET_HDR_LEN);

	pkt_hdr.pkt_id = htonl(SIMPLE_PACKET_ID);
	pkt_hdr.cmd_id = htonl(cmd_num);
	pkt_hdr.pkt_len = htonl(sizeof(struct simple_vp_ips));

	memcpy(trans_buff, &pkt_hdr, SIMPLE_PACKET_HDR_LEN);

	memcpy(&(vp_ips.vp_uip), &(mdevice->vpn_uip), sizeof(struct in6_addr));
	memcpy(&(vp_ips.vp_ip), &(mdevice->vpn_ip), sizeof(struct in6_addr));

	DEBUG_MSG("SIMPLE - Sending vip6: %s and ip6: %s\n", 
			  inet_ntop(AF_INET6, &vp_ips.vp_uip, vip6_disp, INET6_ADDRSTRLEN),
			  inet_ntop(AF_INET6, &vp_ips.vp_ip, ip6_disp, INET6_ADDRSTRLEN));

	memcpy(trans_buff+SIMPLE_PACKET_HDR_LEN, &vp_ips, sizeof(struct simple_vp_ips));

	pkt_len = sizeof(struct simple_vp_ips)+SIMPLE_PACKET_HDR_LEN;

	memcpy(mnode->w_buffer + mnode->w_bytes, trans_buff, pkt_len);
	mnode->w_bytes += pkt_len;

	event_add(&mnode->write_event, NULL);
}
