/*    Copyright (C) 2010  Tomasz Bursztyka
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __VPZONE_H__
#define __VPZONE_H__

#include <stdbool.h>
#include <stdio.h>
#include <net/if.h>
#include <netinet/in.h>
#include <sys/time.h>

#include <ares.h>
#include <event.h>

#define TUN_PKT_LEN 40

#define IP_HOSTNAME_MAX_LENGTH 256

#define CMD_LINE_MAX_LENGTH 1024

#define VPZONE_DEFAULT_PORT 44419

#define COMMAND_NAME_MAX_LENGTH 32

#define COMMAND_MAX_PARAMS 128

#define PLUGIN_NAME_MAX_LENGTH 32

#define PLUGIN_TYPE_NAME_MAX 32

#define DEFAULT_NODE_LIMIT 100

#define VPZONE_DEFAULT_BUFFER_SIZE 16*1024

#define VPZONE_DEFAULT_MAX_DNS_REQ ARES_GETSOCK_MAXNUM

#define VPZONE_MAX_DNS_REQ VPZONE_DEFAULT_MAX_DNS_REQ

#define VPZONE_DEFAULT_TIMEOUT 30

typedef int (*valuecmp_f) (const void *, const void*);

enum _ptable_type_e
{
	PTABLE_FIXED = 0x1,
	PTABLE_TRIMED = 0x10,
	PTABLE_SEARCH = 0x100
};

typedef enum _ptable_type_e ptable_type_t;

struct _ptable_s
{
	void **ptable;
	valuecmp_f value_compare;
	int type;
	int length;
	int count;
};

typedef struct _ptable_s ptable_t;

/**
 * A type definition for data extension in structures.
 * Extents  are maint  to exist  in mnode_t  and mserver_t  structures as
 * extents table.   It strictly depends  on plugins. Each plugin  owns an
 * id, which  id represents  the index in  extents table.   Therefore, on
 * every structures  seen previously,  a plugin can  manage its  own data
 * extension through the given extents table and its id as index.
 */
typedef void * extent_t;

/**
 * MNode data structure.
 * Stores all necessary informations about one mnode, which is created on
 * connection  (connect/accept) and  exists until  its  disconnection. If
 * more informations  needs to be stored,  use its extents  table.  It is
 * maint to be kept in a  hash map in mdevice structure where vpn_uip and
 * vpn_ip are used as keys.  You should use its typedef mnode_t.
 * @see extent_t
 * @see mdevice_t
 */
struct _mnode_s
{
	int sock;

	struct in6_addr vpn_ip;
	struct in6_addr vpn_uip;

	int connector_id;
	int cipher_id;
	int authenticator_id;

	char host[IP_HOSTNAME_MAX_LENGTH];
	uint16_t port;

	struct sockaddr address;

	ptable_t *extents;

	uint64_t bytes_send;
	uint64_t bytes_recv;

	bool is_server;

	int r_bytes;
	uint8_t *r_buffer;

	int w_bytes;
	int w_rests;
	uint8_t *w_buffer;

	struct event read_event;
    struct event write_event;
};

/**
 * A type definition for MNode data structure.
 * This type is preferred against "struct _mnode_s". 
 * @see _mnode_s
 */
typedef struct _mnode_s mnode_t;

/**
 * MServer data structure.
 * Stores all  necessary information about on mserver,  which exists for
 * every server configuration  set (0...n). It is maint to  be kept as a
 * linked list.
 * @see mdevice_t
 */
struct _mserver_s
{
	int sock;

	int connector_id;
	int cipher_id;
	int authenticator_id;

	char host[IP_HOSTNAME_MAX_LENGTH];
	uint16_t port;

	struct sockaddr address;

	ptable_t *extents;

	struct event server_event;

	struct _mserver_s *next;
};

/**
 * A type definition for MServer data structure.
 * This type is preferred against "struct _mserver_s". 
 * @see _mserver_s
 */
typedef struct _mserver_s mserver_t;

/**
 * MDevice data structure.
 * It is the major structure which exists as a singleton. It handles tun
 * device,  the  IO  through   event_base  and  owns  the  mservers  and
 * mnodes. This is only the one who keeps the configuration of the whole
 * software instance.  Therefore it  exists as a  singleton shared  in a
 * global variable.
 * @see mnode_t
 * @see mnode_s
 */
struct _mdevice_s
{
	char config_file[FILENAME_MAX];

	int tun_device;
	char dev_name[IF_NAMESIZE];

	int authenticator_id;

	struct timeval timeout;
	bool prio_ipv4;

	struct in6_addr vpn_ip;
	struct in6_addr vpn_uip;

	ares_channel dns_req_channel;
	ptable_t *dns_reqs;

	struct event_base *base_event;

	struct event read_event;
	struct event write_event;

	int buffer_size;
	int w_bytes;
	uint8_t *w_buffer;

	mserver_t *mservers;

	void *ip_mnodes;

	int mnodes_limit;
};

/**
 * A type definition for MDevice data structure.
 * This type is preferred against "struct _mdevice_s". 
 * @see _mdevice_s
 */
typedef struct _mdevice_s mdevice_t;

/**
 * Logging levels enumeration.
 * Each level  means a verbosity level:  from 0 (only  fatal messages are
 * shown) to 4 (everything and debug messages are shown).
 */
enum _log_level_e
{
	FATAL_LVL,
	ERROR_LVL,
	WARNING_LVL,
	INFO_LVL,
#ifdef DEBUG
	DEBUG_LVL,
#endif
	MAX_LOG_LVL
};

/**
 * A type definition for log levels enumeration.
 * This type is preferred against "enum _log_level_e". 
 * @see _log_level_e
 */
typedef enum _log_level_e log_level_t;

typedef void (*log_msg_f) (log_level_t, const char *, ...);

/**
 * A type definition for handlers.
 * A handler is a function pointer.  A handler is maint to be hooked only
 * once  in the whole  software instance.  There are  a little  amount of
 * handlers  which target important  tasks in  the core  (routing packets
 * routine for instance).
 * @see handler_num_t
 */
typedef void * handler_t;

/**
 * Callback data structure.
 * A callback, on  the contrary with a handler, can  be hooked many times
 * and  exists, therefore,  as a  linked list  of handlers.  This targets
 * small actions  which should  be called before/after  certain strategic
 * procedures   (mnode    accept   for   instance)    but   without   any
 * priorities.  Callbacks  are set  by  plugins,  so  the position  of  a
 * callback in the  list depends on plugin loading  position among others
 * and if other plugins have already hooked the same callback.
 * @see handler_t
 */
struct _callback_s
{
	handler_t callback;
	struct _callback_s *next;
};

/**
 * A type definition for callbacks.
 * This type is preferred against "struct _callback_s".
 * @see _callback_s
 */
typedef struct _callback_s callback_t;

/**
 * Handlers type enumeration.
 * This defines the type of handler you can hook. Names are self-documenting
 */
enum _handler_num_e
{
	/*Simple handlers*/
	HDL_MANAGE_PACKET,
	HDL_ROUTE_PACKET,
	HDL_SELECT_ROUTE,
	HDL_READ_DEVICE,
	HDL_WRITE_DEVICE,
	HDL_REMOVE_MNODE,
	HDL_REMOVE_MSERVER,
	HDL_REMOVE_MDEVICE,
	/*Last: numbers of handlers*/
	HDL_MAX_HANDLERS
};

/**
 * A type definition for handlers type.
 * This type is preferred against "enum _handler_num_e".
 * @see _handler_num_e
 */
typedef enum _handler_num_e handler_num_t;

/**
 * Callbacks type enumeration.
 * This defines the type of callback you can hook. Names are self-documenting
 */
enum _callback_num_e
{
	/*linked list handlers: callbacks.*/
	CB_MANAGE_INTERNAL_PACKET,
	CB_POST_ACCEPT,
	CB_POST_CONNECT,
	CB_PRE_DISCONNECT,
	/*These are callbacks dedicated to extents*/
	CB_INIT_MSERVER_EXTENT,
	CB_REMOVE_MSERVER_EXTENT,
	CB_INIT_MNODE_EXTENT,
	CB_REMOVE_MNODE_EXTENT,
	/*Last: numbers of callbacks*/
	CB_MAX_CALLBACKS
};

/**
 * A type definition for callbacks type.
 * This type is preferred against "enum _callback_num_e".
 * @see _callback_num_e
 */
typedef enum _callback_num_e callback_num_t;

/**
 * Hook Codes.
 * This codes are the ones returned while trying to hook a handler/callback.
 * @see handler_num_t
 * @see callback_num_t
 */
enum _hook_code_e
{
	HOOK_SUCCESSFUL,
	HOOK_ALREADY_SET,
	HOOK_UNKNOWN
};

/**
 * A type definition for hook codes.
 * This type is preferred against "enum _hook_code_e".
 * @see _hook_code_e
 */
typedef enum _hook_code_e hook_code_t;

/**
 * Level of commands.
 */
enum _cmd_level_e
{
	CMD_EXTERNAL,
	CMD_INTERNAL,
	CMD_NONE
};

/**
 * A type definition for command levels.
 * This type is preferred against "enum _cmd_level_e".
 * @see _cmd_level_e
 */
typedef enum _cmd_level_e cmd_level_t;

/**
 * A type definition for command signature.
 * All commands should follow this signature.
 */
typedef int (*command_f) (int, char **);

/**
 * Command data structure.
 * Each command (internal  or external) is stored in  such structure.  it
 * provides the command  name to use, the function related  to it and the
 * (non Mandatory) help message.
 * @see command_f
 */
struct _command_s
{
	char name[COMMAND_NAME_MAX_LENGTH];
	cmd_level_t level;
	command_f command;
	const char *help;
};

/**
 * A type definition for commands.
 * This type is preferred against "struct _command_s".
 * @see _command_s
 */
typedef struct _command_s command_t;

/**
 * Code returned by commands.
 */
enum _cmd_code_e
{
	CMD_SUCCESS,
	CMD_UNKNOWN,
	CMD_ERROR
};

/**
 * A type definition for command codes.
 * This type is preferred against "enum _cmd_code_e".
 * @see _cmd_code_e
 */
typedef enum _cmd_code_e cmd_code_t;

typedef void (*manage_packet_f) (mnode_t *);
typedef int  (*route_packet_f) (uint8_t *, int, void *);
typedef void (*read_tun_f) (int, short, void *);
typedef void (*write_tun_f) (int, short, void *);
typedef void (*remove_mnode_f) (mdevice_t *, mnode_t *);
typedef void (*remove_mserver_f) (mdevice_t *, mserver_t *);
typedef void (*remove_mdevice_f) (void);
typedef mnode_t * (*select_route_f) (struct in6_addr *);
typedef int  (*manage_internal_packet_f) (callback_t *, mdevice_t *, mnode_t *);
typedef void (*post_accept_f) (callback_t *, mdevice_t *, mnode_t *);
typedef void (*post_connect_f) (callback_t *, mdevice_t *, mnode_t *);
typedef void (*pre_disconnect_f) (callback_t *, mnode_t *);
typedef void (*init_mnode_extent_f) (callback_t *, mnode_t *);
typedef void (*remove_mnode_extent_f) (callback_t *, mnode_t *);
typedef void (*init_mserver_extent_f) (callback_t *, mserver_t *);
typedef void (*remove_mserver_extent_f) (callback_t *, mserver_t *);

typedef int  (*connect_mnode_f) (mnode_t *);
typedef void (*disconnect_mnode_f) (mnode_t *);
typedef int  (*setup_mserver_f) (mdevice_t *, mserver_t *);
typedef void (*accept_mnode_f) (int, short, void *);
typedef int  (*close_mserver_f) (mdevice_t *, mserver_t *);

typedef int  (*setup_mnode_server_cipher_f) (mnode_t *);
typedef int  (*setup_mnode_client_cipher_f) (mnode_t *);
typedef int  (*deinit_mnode_server_cipher_f) (mnode_t *);
typedef int  (*deinit_mnode_client_cipher_f) (mnode_t *);
typedef void (*recv_mnode_f) (int, short, void *);
typedef void (*send_mnode_f) (int, short, void *);

typedef int  (*setup_mnode_server_authentification_f) (mnode_t *);
typedef int  (*setup_mnode_client_authentification_f) (mnode_t *);
typedef int  (*deinit_mnode_server_authentification_f) (mnode_t *);
typedef int  (*deinit_mnode_client_authentification_f) (mnode_t *);

/**
 * Connector plugin data structure.
 * A connector is a type of plugin which should provide necessary handler
 * set to provide  client/server connectivity. It is a  linked list which
 * means a plugin may provide many connectors in a row.
 */
struct _connector_s
{
	int plugin_id;

	char type[PLUGIN_TYPE_NAME_MAX];

	connect_mnode_f connect_mnode_h;
	disconnect_mnode_f disconnect_mnode_h;
	setup_mserver_f setup_mserver_h;
	accept_mnode_f accept_mnode_h;
	close_mserver_f close_mserver_h;
};

/**
 * A type definition for connector plugin.
 * This type is preferred against "struct _connector_s"
 * @see _connector_s
 */
typedef struct _connector_s connector_t;

/**
 * A type definition for mandatory connectors access in plugins.
 * Such  function needs  to be  exposed by  connector plugins.  It should
 * return current plugin connectors linked list.
 * @see _connector_s
 */
typedef connector_t * (*get_connectors_f) (void);

/**
 * Cipher plugin data structure.
 * A cipher  is a type of  plugin which should  provide necessary handler
 * set to  provide client/server  ciphering capability while  sending and
 * receiving data.  It is a linked  list which means a plugin may provide
 * many connectors in a row.
 */
struct _cipher_s
{
	int plugin_id;

	char type[PLUGIN_TYPE_NAME_MAX];

	setup_mnode_server_cipher_f setup_mnode_server_cipher_h;
	setup_mnode_client_cipher_f setup_mnode_client_cipher_h;
	deinit_mnode_server_cipher_f deinit_mnode_server_cipher_h;
	deinit_mnode_client_cipher_f deinit_mnode_client_cipher_h;
	recv_mnode_f recv_mnode_h;
	send_mnode_f send_mnode_h;
};

/**
 * A type definition for cipher plugin.
 * This type is preferred against "struct _cipher_s"
 * @see _cipher_s
 */
typedef struct _cipher_s cipher_t;

/**
 * A type definition for mandatory ciphers access in plugins.
 * Such function needs to be  exposed by cipher plugins. It should return
 * current plugin ciphers linked list.
 * @see _cipher_s
 */
typedef cipher_t * (*get_ciphers_f) (void);

/**
 * Authenticator plugin data structure.
 * An authenticator  is a type  of plugin which should  provide necessary
 * handler set  to provide client/server  authentification capability. It
 * is a linked list which  means a plugin may provide many authenticators
 * in a row.
 */
struct _authenticator_s
{
	int plugin_id;

	char type[PLUGIN_TYPE_NAME_MAX];

	setup_mnode_server_authentification_f setup_mnode_server_authentification_h;
	setup_mnode_client_authentification_f setup_mnode_client_authentification_h;
	deinit_mnode_server_authentification_f deinit_mnode_server_authentification_h;
	deinit_mnode_client_authentification_f deinit_mnode_client_authentification_h;
};

/**
 * A type definition for authenticator plugin.
 * This type is preferred against "struct _authenticators_s"
 * @see _authenticators_s
 */
typedef struct _authenticator_s authenticator_t;

/**
 * A type definition for mandatory authenticators access in plugins.
 * Such function needs to be  exposed by authenticator plugins. It should 
 * return current plugin authenticators linked list. 
 * @see _authenticators_s
 */
typedef authenticator_t * (*get_authenticators_f) (void);

/**
 * Plugin type code.
 * These  are the  code returned  by a  plugin when  you call  its init()
 * procedure.   Note that it  might return  any '||'-made  combination of
 * code.  (eg. PLUGIN_CONNECTOR || PLUGIN_CIPHER) This means an important
 * issue: a plugin can provide as many type of service it wants.
 */
enum _plugin_type_e
{
	PLUGIN_DEFAULT = 0x0,
	PLUGIN_CONNECTOR = 0x1,
	PLUGIN_CIPHER = 0x10,
	PLUGIN_AUTHENTICATOR = 0x100,
	MAX_PLUGIN_TYPE
};

/**
 * A type definition for plugin type code.
 * This type is preferred against "enum _plugin_type_e"
 * @see _plugin_type_e
 */
typedef enum _plugin_type_e plugin_type_t;

/**
 * Extent type code.
 * Each plugin type  might set an extent to  mnode/mserver structures. To
 * set  or  retrieve  such  extent,   it  should  use  such  extent  type
 * code.  Extents it  mnode/mserver are  2 dimention  table  of extent_t:
 * first index  targets the plugin through  its id, second  one need such
 * extent type code.
 * @see extent_t
 * @see mnode_t
 * @see mserver_t
 */
enum _extent_type_e
{
	EXTENT_DEFAULT,
	EXTENT_CONNECTOR,
	EXTENT_CIPHER,
	EXTENT_AUTHENTICATOR,
	MAX_EXTENT_TYPE
};

/**
 * A type definition for extent type code.
 * This type is preferred against "enum _extent_type_e"
 * @see _extent_type_e
 */
typedef enum _extent_type_e extent_type_t;

/**
 * A data structure to gather exportable functions into plugins.
 * Such structure exists  as a singleton. It is  shared among all plugins
 * which therefore  can use provided  functions and, the  most important,
 * can manipulate  the mdevice data. It also  provide connectors, ciphers
 * and authenticators ptable.
 * @see mdevice_t
 * @see connector_t
 * @see cipher_t
 * @see authenticator_t
 */
struct _in_plugin_handlers_s
{
	/**
	 * @see log_msg
	 */
	log_msg_f log_msg_h;
	/**
	 * @see log_set_redirection
	 */
	void (*log_set_redirection_h) (log_msg_f);
	/**
	 * @see log_unset_redirection
	 */
	void (*log_unset_redirection_h) (void);
    /**
	 * @see set_handler
	 */
	int (*set_handler_h) (handler_num_t, handler_t);
    /**
	 * @see unset_handler
	 */
	int (*unset_handler_h) (handler_num_t);
    /**
	 * @see add_callback
	 */
	int (*add_callback_h) (callback_num_t, handler_t);
    /**
	 * @see sub_callback
	 */
	int (*sub_callback_h) (callback_num_t, handler_t);
    /**
	 * @see init_mnode
	 */
	mnode_t * (*init_mnode_h) (void);
    /**
	 * @see free_mnode
	 */
	void (*free_mnode_h) (mnode_t *);
     /**
	 * @see pt_create
	 */
	ptable_t * (*pt_create_h) (ptable_type_t, int, valuecmp_f);
     /**
	 * @see pt_insert
	 */
	int (*pt_insert_h) (ptable_t *, void *);
     /**
	 * @see pt_search
	 */
	void * (*pt_search_h) (ptable_t *, void *);
     /**
	 * @see pt_remove
	 */
	void * (*pt_remove_h) (ptable_t *, int);
     /**
	 * @see pt_destroy
	 */
	int (*pt_destroy_h) (ptable_t *);
     /**
	 * @see init_command_set
	 */
	ptable_t * (*init_command_set_h) (const command_t *);
	/**
	 * @see add_command
	 */
	int (*add_command_h) (command_t *);
	/**
	 * @see del_command
	 */
	int (*del_command_h) (command_t *);
	/**
	 * @see parse_command
	 */
	char ** (*parse_command_h) (char *);
    /**
	 * @see parse_configuration
	 */
	void (*parse_configuration_h) (char *, ptable_t *);
	/**
	 * @see interpret_command
	 */
	int (*interpret_command_h) (char **, cmd_level_t, ptable_t *);
	/**
	 * @see free_cmd_line
	 */
	void (*free_cmd_line_h) (char **);
    /**
	 * @see h_create
	 */
	void * (*h_create_h) (uint32_t, uint32_t (*) (void *), int (*) (void *, void*));
    /**
	 * @see h_insert
	 */
	int (*h_insert_h) (void *, void *, void *);
    /**
	 * @see h_search
	 */
	void * (*h_search_h) (void *, void *);
    /**
	 * @see h_remove
	 */
	void * (*h_remove_h) (void *, void *, bool);
    /**
	 * @see h_destroy
	 */
	int (*h_destroy_h) (void *);
    /**
	 * @see h_count
	 */
	uint32_t (*h_count_h) (void *);
	/**
	 * @see h_iter_begin
	 */
	int (*h_iter_begin_h) (void *);
	/**
	 * @see h_iter_end
	 */
	void (*h_iter_end_h) (void *);
	/**
	 * @see h_iter_next
	 */
	bool (*h_iter_next_h) (void *);
	/**
	 * @see h_iter_getkey
	 */
	void * (*h_iter_getkey_h) (void *);
	/**
	 * @see h_iter_getvalue
	 */
	void * (*h_iter_getvalue_h) (void *);
    /**
	 * @see compare_ipv6
	 */
	int (*compare_ipv6_h) (void *, void *);
    /**
	 * @see hash_ipv6
	 */
	uint32_t (*hash_ipv6_h) (void *);
    /**
	 * @see set_extent
	 */
	int (*set_extent_h) (bool, void *, int, extent_type_t, extent_t);
    /**
	 * @see get_extent
	 */
	extent_t (*get_extent_h) (bool, void *, int, extent_type_t);

	handler_t handlers[HDL_MAX_HANDLERS];

	callback_t *callbacks[CB_MAX_CALLBACKS];

	ptable_t *connectors;

	ptable_t *ciphers;

	ptable_t *authenticators;
    /**
	 * @see mdevice_t
	 */
	mdevice_t *mdevice_g;
};

/**
 * A type definition for plugin functions exporter.
 * This type is preferred against "struct _in_plugin_handlers_s"
 * @see _in_plugin_handlers_s
 */
typedef struct _in_plugin_handlers_s in_plugin_handlers_t;

/**
 * A type definition representing an opened plugin.
 * (set  and used  by dlfcnd  function  set. See  man 3  dlopen for  more
 * informations.)
 */
typedef void * plugin_handle_t;

/**
 * A type definition for plugin init() mandatory function.
 */
typedef int (*plugin_init_f) (int, in_plugin_handlers_t *);

/**
 * A type definition for plugin unload() mandatory function.
 */
typedef int (*plugin_unload_f) (void);

/**
 * A data structure to store plugin informations.
 * This is the  way it keeps tracks of loaded  plugins. Storing its name,
 * id, type etc...  as a linked list. The 2  mandatory functions a plugin
 * needs to expose are also kept in memory, just after loading the plugin
 * file.
 */
struct _plugin_s
{
	char name[PLUGIN_NAME_MAX_LENGTH];

	const char * dependencies;

	int id;

	int type;

	plugin_handle_t handle;

	/*Mandatory functions exposed by  plugin*/
	plugin_init_f init;
	plugin_unload_f unload;
};

/**
 * A type definition for plugin storage.
 * This type is preferred against "struct _plugin_s"
 * @see _plugin_s
 */
typedef struct _plugin_s plugin_t;

/*Some useful macros used intenally*/

#define HDL_RET(ret_var, def_val, hdl_f, hdl, ...)		\
	if (hdl != NULL)									\
		ret_var = ((hdl_f)hdl)(__VA_ARGS__);			\
	else												\
		ret_var = def_val;

#define CB_VOID(cb_f, cb, ...) {										\
		if (cb != NULL)													\
			((cb_f)((callback_t *)cb)->callback)(cb, __VA_ARGS__);		\
	}

#define CB_RET(cb_f, cb, ret_var, def_val, ...) {						\
		if (cb != NULL)													\
			ret_var = ((cb_f)((callback_t *)cb)->callback)(cb, __VA_ARGS__); \
		else															\
			ret_var = def_val;											\
	} 

#endif
